/*
 * DomUI Java User Interface library
 * Copyright (c) 2010 by Frits Jalvingh, Itris B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * See the "sponsors" file for a list of supporters.
 *
 * The latest version of DomUI and related code, support and documentation
 * can be found at http://www.domui.org/
 * The contact for the project is Frits Jalvingh <jal@etc.to>.
 */
package to.etc.webapp.qsql;

import java.math.*;
import java.sql.*;
import java.util.*;

import to.etc.util.*;

/**
 * Utility class for JDBC code.
 *
 *
 * <h2>Preparation for calling SP's with TYPE xx IS RECORD parameters.</h2>
 * <p>Datadict table ALL_PROCEDURES contains all SP's in packages. The parameters for
 * SPs can be glanced from ALL_ARGUMENTS; something odd so far is that doing a selection:
 * <pre>
 * select * from sys.all_arguments where owner='DECADE' and package_name='GEBRUI' and object_name='LEES100';
 * </pre>
 * returns data that seem to indicate that the SP exists as a 2-parameter version but also an expanded version
 * having all parameter fields.</p>
 *
 * @author <a href="mailto:jal@etc.to">Frits Jalvingh</a>
 * Created on Dec 21, 2009
 */
public class JdbcUtil {
	private JdbcUtil() {}

	static public void setLong(PreparedStatement ps, int index, Long value) throws SQLException {
		if(value == null)
			ps.setNull(index, Types.NUMERIC);
		else
			ps.setLong(index, value.longValue());
	}

	static public void setInteger(PreparedStatement ps, int index, Integer value) throws SQLException {
		if(value == null)
			ps.setNull(index, Types.NUMERIC);
		else
			ps.setInt(index, value.intValue());
	}

	static public void setDouble(PreparedStatement ps, int index, Double value) throws SQLException {
		if(value == null)
			ps.setNull(index, Types.DOUBLE);
		else
			ps.setDouble(index, value.doubleValue());
	}

	/**
	 * Sets a TIMESTAMP value containing both TIME and DATE values.
	 * @param ps
	 * @param index
	 * @param value
	 * @throws SQLException
	 */
	static public void setTimestamp(PreparedStatement ps, int index, java.util.Date value) throws SQLException {
		if(value == null)
			ps.setNull(index, Types.TIMESTAMP);
		else if(value instanceof Timestamp)
			ps.setTimestamp(index, (Timestamp) value);
		else
			ps.setTimestamp(index, new Timestamp(value.getTime()));
	}

	/**
	 * Sets a <b>truncated</b> date containing <i>only</i> the date part and a zero time.
	 * @param ps
	 * @param index
	 * @param value
	 * @throws SQLException
	 */
	static public void setDateTruncated(PreparedStatement ps, int index, java.util.Date value) throws SQLException {
		if(value == null)
			ps.setNull(index, Types.TIMESTAMP);
		else {
			value = DateUtil.truncateDate(value);
			ps.setDate(index, new java.sql.Date(value.getTime()));
		}
	}

	static public void	setString(PreparedStatement ps, int index, String value) throws SQLException {
		if(value == null || value.trim().length() == 0)
			ps.setNull(index, Types.VARCHAR);
		else
			ps.setString(index, value);
	}

	static public void setStringTruncated(PreparedStatement ps, int index, String value, int maxlen) throws SQLException {
		if(value == null || value.trim().length() == 0)
			ps.setNull(index, Types.VARCHAR);
		else {
			int len = value.length();
			if(len > maxlen)
				value = value.substring(0, maxlen);
			ps.setString(index, value);
		}
	}

	static public void setYN(PreparedStatement ps, int index, Boolean value) throws SQLException {
		if(value == null)
			ps.setNull(index, Types.VARCHAR);
		else
			ps.setString(index, value.booleanValue() ? "Y" : "N");
	}

	static public void setFK(PreparedStatement ps, int index, ILongIdentifyable foreigner) throws SQLException {
		if(foreigner == null)
			ps.setNull(index, Types.NUMERIC);
		else if(foreigner.getId() == null)
			throw new IllegalStateException("Reference to foreign object '" + foreigner + "' has a null ID.");
		else {
			ps.setLong(index, foreigner.getId().longValue());
		}
	}

	/**
	 * Quick method to select a single value of a given type from the database. Returns null if not found AND if the value was null...
	 * @param connection
	 * @param clz
	 * @param select
	 * @return
	 */
	public static <T> T selectOne(Connection connection, Class<T> clz, String select, Object... params) throws SQLException {
		PreparedStatement ps = null;
		ResultSet rs = null;
		try {
			ps = connection.prepareStatement(select);
			setParameters(ps, 1, params);
			rs = ps.executeQuery();
			if(! rs.next())
				return null;
			Object res = null;
			if(clz == String.class)
				return (T) rs.getString(1);
			else if(clz == Long.class || clz == long.class)
				res = Long.valueOf(rs.getLong(1));
			else if(clz == Integer.class || clz == int.class)
				res = Integer.valueOf(rs.getInt(1));
			else if(clz == java.util.Date.class) {
				java.sql.Timestamp ts = rs.getTimestamp(1);
				if(ts != null)
					res = new java.util.Date(ts.getTime());
			} else if(clz == Boolean.class || clz == boolean.class)
				res = Boolean.valueOf(rs.getBoolean(1));
			else
				throw new IllegalStateException("Call error: cannot handle requested return type " + clz);
			if(rs.wasNull())
				return null;
			return (T) res;
		} catch(SQLException x) {
			String msg = x.getMessage();
			if(rs != null && msg != null && msg.contains("internal representation")) {
				String res = "(cannot obtain value)";
				try {
					res = rs.getString(1);
				} catch(Exception xx) {}
				throw new SQLException("Cannot convert '" + res + "' to internal representation " + clz + ": " + x, x);
			}
			throw x;
		} finally {
			try {
				if(rs != null)
					rs.close();
			} catch(Exception x) {}
			try {
				if(ps != null)
					ps.close();
			} catch(Exception x) {}
		}
	}

	/**
	 * Quick method to select a single value of a given type from the database. Returns null if not found AND if the value was null...
	 * @param connection
	 * @param clz
	 * @param select
	 * @return
	 */
	public static <T> List<T> selectSingleColumnList(Connection connection, Class<T> clz, String select, Object... params) throws SQLException {
		PreparedStatement ps = null;
		ResultSet rs = null;
		try {
			ps = connection.prepareStatement(select);
			setParameters(ps, 1, params);
			rs = ps.executeQuery();
			List<T> list = new ArrayList<T>();
			while(rs.next()) {
				Object res = null;
				if(clz == String.class)
					res = rs.getString(1);
				else if(clz == Long.class || clz == long.class)
					res = Long.valueOf(rs.getLong(1));
				else if(clz == Integer.class || clz == int.class)
					res = Integer.valueOf(rs.getInt(1));
				else if(clz == java.util.Date.class) {
					java.sql.Timestamp ts = rs.getTimestamp(1);
					if(ts != null)
						res = new java.util.Date(ts.getTime());
				} else if(clz == Boolean.class || clz == boolean.class)
					res = Boolean.valueOf(rs.getBoolean(1));
				else
					throw new IllegalStateException("Call error: cannot handle requested return type " + clz);
				if(rs.wasNull())
					list.add(null);
				else
					list.add((T) res);
			}
			return list;
		} catch(SQLException x) {
			String msg = x.getMessage();
			if(rs != null && msg != null && msg.contains("internal representation")) {
				String res = "(cannot obtain value)";
				try {
					res = rs.getString(1);
				} catch(Exception xx) {}
				throw new SQLException("Cannot convert '" + res + "' to internal representation " + clz + ": " + x, x);
			}
			throw x;
		} finally {
			try {
				if(rs != null)
					rs.close();
			} catch(Exception x) {}
			try {
				if(ps != null)
					ps.close();
			} catch(Exception x) {}
		}
	}

	public static void setParameters(PreparedStatement ps, int startindex, Object[] params) throws SQLException {
		if(params == null)
			return;
		for(int i = 0; i < params.length; i++) {
			Object val = params[i];
			int px = i + startindex;
			if(val instanceof JdbcOutParam< ? >) {
				if(!(ps instanceof CallableStatement)) {
					throw new IllegalArgumentException("expected CallableStatement instead of PreparedStatement for OUT/IN OUT param function/procedure call!");
				}
				((CallableStatement) ps).registerOutParameter(px, calcSQLTypeFor(((JdbcOutParam< ? >) val).getClassType()));
				if(val instanceof JdbcInOutParam< ? >) {
					setParameter(ps, ((JdbcInOutParam< ? >) val).getValue(), px);
				}
			} else {
				setParameter(ps, val, px);
			}
		}
	}

	private static void setParameter(PreparedStatement ps, Object val, int px) throws SQLException {
		if(val == null)
			ps.setString(px, null);
		else if(val instanceof String) {
			ps.setString(px, ((String) val));
		} else if(val instanceof Long) {
			ps.setLong(px, ((Long) val).longValue());
		} else if(val instanceof Integer) {
			ps.setInt(px, ((Integer) val).intValue());
		} else if(val instanceof BigDecimal) {
			ps.setBigDecimal(px, (BigDecimal) val);
		} else if(val instanceof Double) {
			ps.setDouble(px, ((Double) val).doubleValue());
		} else if(val instanceof java.sql.Timestamp) {
			ps.setTimestamp(px, (java.sql.Timestamp) val);
		} else if(val instanceof java.util.Date) {
			ps.setTimestamp(px, new Timestamp(((java.util.Date) val).getTime()));
		} else if(val instanceof Boolean) {
			ps.setBoolean(px, ((Boolean) val).booleanValue());
		} else
			throw new IllegalStateException("Call error: unknown SQL parameter of type " + val.getClass());
	}

	public static void readParameters(PreparedStatement ps, int startindex, Object[] params) throws SQLException {
		if(params == null)
			return;
		for(int i = 0; i < params.length; i++) {
			Object val = params[i];
			if(val instanceof JdbcOutParam< ? >) {
				if(!(ps instanceof CallableStatement)) {
					throw new IllegalArgumentException("expected CallableStatement instead of PreparedStatement for OUT/IN OUT param function/procedure call!");
				}
				setOutParamValue((CallableStatement) ps, startindex + i, ((JdbcOutParam< ? >) val).getClassType(), (JdbcOutParam< ? >) val);
			}
		}
	}

	static public List<JdbcAnyRecord> queryAny(Connection dbc, String select, Object... parameters) throws SQLException {
		PreparedStatement ps = null;
		ResultSet rs = null;
		try {
			ps = dbc.prepareStatement(select);
			setParameters(ps, 1, parameters);
			rs = ps.executeQuery();
			return queryAny(select, rs);
		} finally {
			try {
				if(rs != null)
					rs.close();
			} catch(Exception x) {}
			try {
				if(ps != null)
					ps.close();
			} catch(Exception x) {}
		}
	}

	static public List<JdbcAnyRecord> queryAny(String tblname, ResultSet rs) throws SQLException {
		List<JdbcAnyRecord> l = new ArrayList<JdbcAnyRecord>();
		ResultSetMetaData md = rs.getMetaData();
		while(rs.next()) {
			JdbcAnyRecord a = new JdbcAnyRecord();
			a.initFromRS(tblname, md, rs);
			l.add(a);
		}
		return l;
	}

	static public JdbcAnyRecord queryAnyOne(Connection dbc, String select, Object... parameters) throws SQLException {
		PreparedStatement ps = null;
		ResultSet rs = null;
		try {
			ps = dbc.prepareStatement(select);
			setParameters(ps, 1, parameters);
			rs = ps.executeQuery();
			return queryAnyOne(select, rs);
		} finally {
			try {
				if(rs != null)
					rs.close();
			} catch(Exception x) {}
			try {
				if(ps != null)
					ps.close();
			} catch(Exception x) {}
		}
	}

	static public JdbcAnyRecord queryAnyOne(String tblname, ResultSet rs) throws SQLException {
		if(!rs.next())
			return null;
		ResultSetMetaData md = rs.getMetaData();
		JdbcAnyRecord a = new JdbcAnyRecord();
		a.initFromRS(tblname, md, rs);
		if(rs.next())
			throw new SQLException("Got >1 result for queryAnyOne");
		return a;
	}

	static private int calcSQLTypeFor(Class< ? > rt) {
		if(rt == String.class)
			return Types.VARCHAR;
		else if(rt == Integer.class || rt == int.class || rt == Long.class || rt == long.class || rt == BigDecimal.class || rt == Double.class || rt == double.class)
			return Types.NUMERIC;
		else
			throw new IllegalStateException("Call error: cannot get SQLType for java type=" + rt);
	}

	private static void appendSPParameters(StringBuilder sb, List<Object> pars, Object[] args) {
		//-- Handle parameters, and handle boolean arguments, sigh.
		for(int i = 0; i < args.length; i++) {
			Object val = args[i];
			if(i > 0)
				sb.append(',');
			if(val instanceof Boolean) {
				sb.append(((Boolean) val).booleanValue() ? "true" : "false");
			} else {
				sb.append("?");
				pars.add(val);
			}
		}
	}

	public static boolean executeStatement(Connection dbc, String sql, Object... args) throws SQLException {
		PreparedStatement ps = null;
		try {
			ps = dbc.prepareStatement(sql);
			setParameters(ps, 1, args);
			return ps.execute();
		} finally {
			try {
				if(ps != null)
					ps.close();
			} catch(Exception x) {}
		}
	}

	/**
	 * Provides interface to call Oracle stored procedures and functions.
	 * In case that called function returns boolean use {@link JdbcUtil#oracleSpCallReturningBool(Connection, String, Object...)}
	 * Support all three type of parameters:<BR/>
	 * <UL>
	 * <LI>OUT params: use {@link JdbcOutParam}</LI>
	 * <LI>IN OUT params: use {@link JdbcInOutParam}</LI>
	 * <LI>IN params: use simple java type instancies</LI>
	 * </UL>
	 * @param <T> Oracle function return value type
	 * @param con Db connection
	 * @param rtype Oracle function return type
	 * @param sp Stored procedure / function name
	 * @param args Stored procedure / function parameters
	 * @return Oracle function return value in case of rtype != Void.class
	 * @throws SQLException
	 */
	static public <T> T oracleSpCall(Connection con, Class<T> rtype, String sp, Object... args) throws SQLException {
		if(rtype == Boolean.class || rtype == boolean.class) {
			return (T) Boolean.valueOf(oracleSpCallReturningBool(con, sp, args));
		}

		StringBuilder sb = new StringBuilder();
		sb.append("begin ");
		int startix = 1;
		if(rtype != null && rtype != Void.class) {
			sb.append("? := ");
			startix = 2;
		}
		List<Object> pars = new ArrayList<Object>(args.length);
		sb.append(sp).append('(');
		appendSPParameters(sb, pars, args);
		sb.append(");");
		sb.append("end;");
		String stmt = sb.toString();
		System.out.println("CALLING: " + stmt);

		//-- Call the SP
		CallableStatement ps = null;
		try {
			ps = con.prepareCall(stmt);
			if(startix != 1)
				ps.registerOutParameter(1, calcSQLTypeFor(rtype));
			setParameters(ps, startix, pars.toArray());

			ps.execute();
			readParameters(ps, startix, args);

			if(startix != 1) {
				return readPsValue(ps, 1, rtype);
			} else {
				return null;
			}
		} finally {
			FileTool.closeAll(ps);
		}
	}

	/**
	 * Similar as {@link JdbcUtil#oracleSpCall(Connection, Class, String, Object...)},
	 * adjusted to handle returning of oracle boolean type properly.
	 * @param con
	 * @param sp
	 * @param args For hanlding IN/OUT/IN OUT params see {@link JdbcUtil#oracleSpCall(Connection, Class, String, Object...)}
	 * @return
	 * @throws SQLException
	 */
	public static boolean oracleSpCallReturningBool(Connection con, String sp, Object... args) throws SQLException {
		StringBuilder sb = new StringBuilder();
		sb.append("declare l_res number; ");
		sb.append("begin if(").append(sp).append("(");
		List<Object> pars = new ArrayList<Object>(args.length);
		appendSPParameters(sb, pars, args);
		sb.append(")) then l_res := 1; else l_res := 0; end if; ? := l_res;");
		sb.append("end;");
		String stmt = sb.toString();
		System.out.println("CALLING: " + stmt + ", outResult=" + (pars.size() + 1));

		//-- Call the SP
		CallableStatement ps = null;
		try {
			ps = con.prepareCall(stmt);
			setParameters(ps, 1, pars.toArray());
			ps.registerOutParameter(pars.size() + 1, Types.NUMERIC);
			ps.execute();
			readParameters(ps, 1, pars.toArray());
			int res = ps.getInt(pars.size() + 1);
			return res != 0;
		} finally {
			FileTool.closeAll(ps);
		}
	}

	private static <T> void setOutParamValue(CallableStatement ps, int index, Class<T> rtype, JdbcOutParam< ? > pOutParam) throws SQLException {
		JdbcOutParam<T> outParam = (JdbcOutParam<T>) pOutParam;
		outParam.setValue(readPsValue(ps, index, rtype));
	}

	private static <T> T readPsValue(CallableStatement ps, int index, Class<T> rtype) throws SQLException {
		if(rtype == String.class) {
			return (T) ps.getString(index);
		} else if(rtype == Integer.class || rtype == int.class) {
			return (T) (Integer.valueOf(ps.getInt(index)));
		} else if(rtype == Long.class || rtype == long.class) {
			return (T) (Long.valueOf(ps.getLong(index)));
		} else if(rtype == Double.class || rtype == double.class) {
			return (T) (Double.valueOf(ps.getDouble(index)));
		} else if(rtype == BigDecimal.class) {
			return (T) ps.getBigDecimal(index);
		} else {
			throw new IllegalStateException("Call error: cannot get out parameter for result java type=" + rtype);
		}
	}

}
